import io
import yamlsettings

from . import BaseObject


# region Settings
class Settings(BaseObject):
    _filename = None

    DEFAULTS = u''

    def __init__(self, filename=None, *args, **kwargs):
        super(Settings, self).__init__(*args, **kwargs)
        self._filename = filename

    def get_settings(self):
        conf = yamlsettings.yamldict.load(io.StringIO(self.DEFAULTS))
        if self._filename is not None:
            yamlsettings.update_from_file(conf, self._filename)

        return conf
# endregion


__all__ = ['Settings']



import io
import sys

import logging
import logging.handlers

from yamlsettings.yamldict import YAMLDict, load

from . import BaseObject
from .utils import Utils


# region Mixins
class CLHMixin(BaseObject):
    DEFAULT_CONF = None

    _conf = None
    _cconf = None

    @property
    def rconf(self):
        """Root config
        """
        return self._conf

    @property
    def gconf(self):
        return self._conf['global'] if self._conf and 'global' in self._conf \
            else self._conf['g'] if self._conf and 'g' in self._conf else None

    @property
    def conf(self):
        return self._cconf

    @property
    def logger(self):
        return self._logger

    def __init__(self, conf=None, logger=None, *args, **kwargs):
        self._conf = conf

        def_conf = YAMLDict()

        if isinstance(self.DEFAULT_CONF, unicode):
            def_conf = load(io.StringIO(self.DEFAULT_CONF))
        elif isinstance(self.DEFAULT_CONF, dict):
            def_conf.update(self.DEFAULT_CONF)

        def iter_over_bases(cls):
            for t in cls.__bases__:
                if t is CLHMixin:
                    return

                yield t.__name__

                for b in iter_over_bases(t):
                    yield b

        if conf is not None:
            # Try to predict config
            for class_name in reversed(list([self.__class__.__name__]) + list(iter_over_bases(self.__class__))):
                names = Utils.camel2underscore(class_name)

                possible_conf_keys = list([names])

                for x in range(1, len(names)):
                    possible_conf_keys.append(names[:-x])

                for pc in possible_conf_keys:
                    if ''.join(pc) in self._conf:
                        def_conf.update(self._conf[''.join(pc)])
                        break

                    if '_'.join(pc) in self._conf:
                        def_conf.update(self._conf['_'.join(pc)])
                        break

        self._cconf = def_conf

        if logger is None:
            try:
                level = self.gconf.logging.level
            except (AttributeError, TypeError):
                level = 'INFO'

            try:
                fmt = self.gconf.logging.format
            except (AttributeError, TypeError):
                fmt = '%(asctime)s %(module)s: %(message)s'

            logger = logging.getLogger('{module}.{class_name}'.format(module=self.__class__.__module__,
                                                                      class_name=self.__class__.__name__))

            if logger.level == logging.NOTSET:
                logger.setLevel(level)

                formatter = logging.Formatter(fmt,
                                              datefmt='%b %d %H:%M:%S')

                try:
                    for h in self.gconf.logging.handlers:
                        if h == 'stdout':
                            handler = logging.StreamHandler(sys.stdout)
                        elif h == 'stderr':
                            handler = logging.StreamHandler(sys.stderr)
                        elif h == 'syslog':
                            handler = logging.handlers.SysLogHandler(address='/dev/log',
                                                                     facility=logging.handlers.SysLogHandler.LOG_DAEMON)

                        logger.addHandler(handler)
                        handler.setFormatter(formatter)

                except (AttributeError, TypeError):
                    handler = logging.StreamHandler(sys.stdout)

                    logger.addHandler(handler)
                    handler.setFormatter(formatter)

        self._logger = logger

        super(CLHMixin, self).__init__(*args, **kwargs)
# endregion


__all__ = ['CLHMixin']



import os
import hashlib

from subprocess import Popen, PIPE

from .mixins import CLHMixin
from .exceptions import PopenError


# region Helpers
class PopenHelper(CLHMixin):
    @staticmethod
    def execute_command(command):
        p = Popen(['/bin/bash',
                   '-eo',
                   'pipefail',
                   '-c',
                   command], stdout=PIPE, stderr=PIPE, close_fds=False)

        stdout, stderr = p.communicate()

        if p.returncode != 0:
            raise PopenError('Command failed: {errormsg}'.format(errormsg=stderr.rstrip('\n')))

        return p.returncode, stdout, stderr


class SshHelper(CLHMixin):
    DEFAULT_CONF = dict(
        ssh_pipelining=False,
        control_persist=1800,
        gzip_upload=True,
        strict_hostkey_checking=False,
        user_known_hosts_file='/dev/null'
    )

    _ssh_opts = str()

    @property
    def ssh_opts(self):
        return self._ssh_opts

    def __init__(self, hostname, *args, **kwargs):
        super(SshHelper, self).__init__(*args, **kwargs)
        self._hostname = hostname
        self._ph = PopenHelper(conf=self.rconf, logger=self.logger)

        if not self.conf.strict_hostkey_checking:
            self._ssh_opts += '-o StrictHostKeyChecking=no '

        self._ssh_opts += '-o UserKnownHostsFile={filename} '.format(
                            filename=self.conf.user_known_hosts_file)

        if self.conf.ssh_pipelining:
            md5 = hashlib.md5()
            md5.update(self._hostname)

            self._ssh_opts += '-oControlMaster=auto ' \
                              '-oControlPersist={control_persist} ' \
                              "-oControlPath='/tmp/ssh-mux-{hhash}%r%p' ".format(
                                  control_persist=self.conf.control_persist,
                                  hhash=md5.hexdigest()
                              )

    def execute_command(self, command):
        return self._ph.execute_command('ssh '
                                        '{ssh_opts} '
                                        "{host} '{command}'".format(
                                                        ssh_opts=self._ssh_opts,
                                                        host=self._hostname,
                                                        command=command
                                        ))

    def putc(self, command, remote_path, gzip=None, gunzip=None):
        if gzip is None and gunzip is None:
            gzip = gunzip = self.conf.gzip_upload

        return self._ph.execute_command('{command} | {gzip_command}'
                                        'ssh {host} '
                                        '{ssh_opts} '
                                        "'{cat_command} > {remote_path}'".format(
                                                            ssh_opts=self._ssh_opts,
                                                            command=command,
                                                            gzip_command='gzip --best |' if gzip else '',
                                                            host=self._hostname,
                                                            remote_path=remote_path,
                                                            cat_command='zcat' if gunzip else 'cat'
                                        ))

    def put(self, path, remote_path, gzip=None, gunzip=None):
        if gzip is None and gunzip is None:
            gzip = gunzip = self.conf.gzip_upload

        if gzip or gunzip:
            return self.putc('cat {path}'.format(path=path), remote_path, gzip, gunzip)

        return self._ph.execute_command('scp -C '
                                        '{ssh_opts} '
                                        "{path} {host}:{remote_path}".format(
                                                                    ssh_opts=self._ssh_opts,
                                                                    host=self._hostname,
                                                                    path=path,
                                                                    remote_path=remote_path
                                        ))

    def is_dir(self, path):
        ret, _, _ = self.execute_command('test -d {path}'.format(
                                                        path=path
                                      ))
        return ret == 0

    def exists(self, path):
        ret, _, _ = self.execute_command('test -e {path}'.format(
                                          path=path
                                      ))
        return ret == 0

    def mkdir(self, path, mode=0o700, recursive=True):
        ret, _, _ = self.execute_command('if [ ! -d {path} ]; then '
                                         'mkdir -p {path}; '
                                         'fi; {chmod_command}'.format(
                                             path=path,
                                             chmod_command='find {path}/ -type d -print0 |'
                                                           'xargs -0 chmod {mode}; '.format(
                                                                 path=os.path.split(path)[0],
                                                                 mode=oct(mode)
                                                           )
                                             if recursive else 'chmod {mode} {path}'.format(
                                                                         path=path,
                                                                         mode=oct(mode)
                                                               )
                                         ))

        return ret == 0

    def rmtree(self, path):
        ret, _, _ = self.execute_command('if [ -d {path} ]; then '
                                         'rm -rf {path}; '
                                         'fi'.format(path=path))

        return ret == 0
# endregion


__all__ = ['PopenHelper', 'PopenError', 'SshHelper']



import os
import re

from multiprocessing import cpu_count

from . import BaseObject


# region Utils
class Utils(BaseObject):
    @staticmethod
    def is_process_running(pid):
        """ Check for the existence of a unix pid. """
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    @staticmethod
    def camel2underscore(camel_input, lower=True):
        words = re.findall(r'[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+', camel_input)
        return map(str.lower, words) if lower else words

    @staticmethod
    def intstringtoseconds(interval_string):
        """Convert a string expressing an interval (e.g. "4D2s") to seconds"""

        _interval_conv_dict = dict(
            s=1,
            m=60,
            h=3600,
            D=86400,
            W=7 * 86400,
            M=30 * 86400,
            Y=365 * 86400
        )

        _interval_regexp = re.compile("^([0-9]+)([smhDWMY])")

        def error():
            raise Exception('Wrong time spec: {interval_string}'.format(
                interval_string=interval_string)
            )

        if len(interval_string) < 2:
            error()

        total = 0
        while interval_string:
            match = _interval_regexp.match(interval_string)
            if not match:
                error()
            num, ext = int(match.group(1)), match.group(2)
            if ext not in _interval_conv_dict or num < 0:
                error()
            total += num * _interval_conv_dict[ext]
            interval_string = interval_string[match.end(0):]
        return total

    @staticmethod
    def cpu_count():
        """Return cpu count"""
        return cpu_count()

    @staticmethod
    def disk_usage(path):
        """Return disk usage in bytes
        """
        st = os.statvfs(path)
        free = st.f_bavail * st.f_frsize
        total = st.f_blocks * st.f_frsize
        used = (st.f_blocks - st.f_bfree) * st.f_frsize
        return int(total), int(used), int(free)
# endregion


__all__ = ['Utils']
